// Copyright (C) 2008
// by all contributors..

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.


#ifndef SCNSL__CORE__PACKET_T_HH
#define SCNSL__CORE__PACKET_T_HH



/// @file
/// A generic packet.

#include <stdexcept>
#include <iostream>

#include "data_types.hh"


namespace Scnsl { namespace Core {

  class Task_if_t;
  class TaskProxy_if_t;
  class Channel_if_t;
  class Node_t;

} }

namespace Scnsl { namespace Core {

  /// @brief A generic packet.
  ///
  /// By default, all the sizes are considered in bits.
  /// The passed buffer is copyed, and not aliased.
  ///
  /// By default a new created packet is valid, has a new id,
  /// has the default printing method.
  ///
  /// Design patterns:
  /// - Regular.
  ///
  class Packet_t
  {
  public:

      /// @name Traits.
      //@{

      /// @brief The byte type.
      typedef Scnsl::Core::byte_t byte_t;

      /// @brief The size type.
      typedef Scnsl::Core::size_t size_t;

      /// @brief The size type.
      typedef Scnsl::Core::counter_t counter_t;

      /// @brief The packet id type.
      typedef Scnsl::Core::packet_id_t packet_id_t;

      /// @brief The packet label type.
      typedef Scnsl::Core::label_t label_t;

      /// @brief The node properties.
      typedef Scnsl::Core::node_properties_t node_properties_t;

      //@}

      /// @brief Default constructor.
      Packet_t()
          throw ();

      /// @brief Copy constructor.
      /// @throw std::bad_alloc If no enough memory is available.
      Packet_t( const Packet_t & p )
          throw ( std::bad_alloc );

      /// @brief Destructor.
      ~Packet_t()
          throw ();

      /// @brief Assignment operator.
      /// @throw std::bad_alloc If no enough memory is available.
      Packet_t & operator = ( const Packet_t & p )
          throw ( std::bad_alloc );


      /// @brief Payload related methods.
      //@{

      /// @brief Checks if payload is stored as packet or as buffer.
      ///
      /// @return True if payload is stored as buffer.
      ///
      bool hasBufferPayload()
          const throw ();

      /// @brief Gets the payload size in bits.
      ///
      /// @return The size.
      ///
      size_t getPayloadSize()
          const throw ();

      /// @brief Gets the payload size in bytes.
      ///
      /// @return The size.
      ///
      size_t getPayloadSizeInBytes()
          const throw ();

      /// @brief Sets the buffer as payload.
      /// The buffer is copied and not aliased.
      ///
      /// @param b The buffer.
      /// @param s The size in bits.
      /// @throw std::bad_alloc If no enough memory is available.
      ///
      void setPayload( const byte_t * b, const size_t s )
          throw ( std::bad_alloc );

      /// @brief Sets the buffer as payload.
      /// The buffer is copied and not aliased.
      ///
      /// @param b The buffer.
      /// @param s The size in bytes.
      /// @throw std::bad_alloc If no enough memory is available.
      ///
      void setPayloadInBytes( const byte_t * b, const size_t s )
          throw ( std::bad_alloc );

      /// @brief Encapsulates a packet.
      /// The encapsulated packet is copyed.
      ///
      /// @param p The packet to encapsulate.
      ///
      void setPayload( const Packet_t & p )
          throw ( std::bad_alloc );

      /// @brief Gets the payload, as a buffer.
      ///
      /// @param b The buffer.
      ///
      void getPayload( byte_t const * & b )
          const throw ();

      /// @brief Gets the payload, as a buffer.
      ///
      /// @param b The buffer.
      ///
      void getPayload( byte_t * & b )
          throw ();

      /// @brief Gets the payload, as a packet.
      ///
      /// @param p The packet.
      ///
      void getPayload( Packet_t & p )
          const throw ();

      /// @brief Gets the payload, as a packet.
      ///
      Packet_t* getInnerPayload( )
          const throw ();

      //@}

      /// @name Header methods.
      //@{

      /// @brief Gets the header size in bits.
      ///
      /// @return The size.
      ///
      size_t getHeaderSize()
          const throw ();

      /// @brief Gets the header size in bytes.
      ///
      /// @return The size.
      ///
      size_t getHeaderSizeInBytes()
          const throw ();

      /// @brief Sets the header size in bits.
      ///
      /// @param s The size.
      ///
      void setHeaderSize( const size_t s )
          throw ();

      /// @brief Sets the header size in bytes.
      ///
      /// @param s The size.
      ///
      void setHeaderSizeInBytes( const size_t s )
          throw ();


      /// @brief Gets the ack flag.
      ///
      /// @return True if this packet is an ack.
      ///
      bool isAck()
          const throw ();

      /// @brief Sets the ack flag.
      ///
      /// @param ack True if this packet is an ack.
      ///
      void setAck( const bool ack )
          throw ();

      /// @brief Gets if this packet requires an ack as reply.
      ///
      /// @return True if an ack is required.
      ///
      bool isAckRequired()
          const throw ();

      /// @brief Sets if this packet requires an ack as reply.
      ///
      /// @param ack True if an ack is required.
      ///
      void setAckRequired( const bool ack )
          throw ();

      /// @brief Gets this packet sequence number.
      ///
      /// @return The sequence number.
      ///
      counter_t getSequenceNumber()
          const throw ();

      /// @brief Sets this packet sequence number.
      ///
      /// @param seqno The sequence number.
      ///
      void setSequenceNumber( const counter_t seqno )
          throw ();

      /// @brief Gets this packet label.
      ///
      /// @return The label.
      ///
      label_t getLabel()
          const throw ();

      /// @brief Sets this packet label.
      ///
      /// @param l The label.
      ///
      void setLabel( const label_t l )
          throw ();


      /// @brief Sets the packet taskproxy sender.
      void setSourceTaskProxy( const TaskProxy_if_t * tp )
          throw ();

      /// @brief Gets the packet taskproxy sender.
      const TaskProxy_if_t * getSourceTaskProxy()
          const throw ();

      /// @brief Sets the packet task destination.
      void setDestinationTask( const Task_if_t * t )
          throw ();

      /// @brief Gets the packet task destination.
      const Task_if_t * getDestinationTask()
          const throw ();


      /// @brief Sets the packet source node.
      void setSourceNode( const Node_t * n )
          throw ();

      /// @brief Gets the packet source node.
      const Node_t * getSourceNode()
          const throw ();

      /// @brief Sets the packet destination node.
      /// NULL means broadcast.
      void setDestinationNode( const Node_t * n )
          throw ();

      /// @brief Gets the packet destination node.
      const Node_t * getDestinationNode()
          const throw ();

      //@}


      /// @name General methods.
      //@{

      /// @brief Gets the total packet size in bits.
      ///
      /// @return The size.
      ///
      size_t getSize()
          const throw ();

      /// @brief Gets the total packet size in bytes.
      ///
      /// @return The size.
      ///
      size_t getSizeInBytes()
          const throw ();

      //@}


      /// @name Simulation methods.
      //@{

      /// @brief Gets the packet id.
      packet_id_t getId()
          const throw ();

      /// @brief Sets the packet id.
      void setId( const packet_id_t id )
          throw ();

      /// @brief Sets the packet propagation channel.
      void setChannel( const Channel_if_t * ch )
          throw ();

      /// @brief Gets the packet propagation channel.
      const Channel_if_t * getChannel()
          const throw ();

      void setSourceProperties(node_properties_t & sp);


      const node_properties_t & getSourceProperties()
		  const;




      //@}


      /// @name Support methods.
      //@{

      /// @brief Equality operator.
      /// We do not consider mixed payload types,
      /// nor simulation fields,
      /// nor fields that will be removed in the future.
      bool operator == ( const Packet_t & p )
          const throw ();

      //@}


      /// @name To be removed methods.
      //@{

      /// @brief Gets the packet validity.
      bool isValid()
          const throw ();

      /// @brief Sets the packet validity.
      void setValidity( const bool v )
          throw ();

      //@}

      /// @brief Static support methods.
      //@{

      /// @brief Gets a new packet id.
      static
      packet_id_t get_new_packet_id()
          throw ();


      /// @brief Gets the required bytes to encode the passed bits.
      static
      size_t to_bytes( const size_t s )
          throw ();

      /// @brief Gets the required bits to encode the passed bytes.
      static
      size_t to_bits( const size_t s )
          throw ();

      //@}

  protected:

      /// @name Payload fields.
      //@{

      /// @brief The internal buffer.
      byte_t * _buffer;

      /// @brief The buffer size in bitsw.
      size_t _buffer_size;

      /// @brief The buffer capacity, in bytes.
      size_t _buffer_capacity;

      Packet_t * _packet;

      //@}


      /// @brief Header fields.
      //@{

      /// @brief True if the packet is an ack.
      bool _is_ack;

      /// @brief True if this packet requires an ack as reply.
      bool _is_ack_required;

      /// @brief This packet sequence number.
      counter_t _sequence_number;

      /// @brief Label of the packet.
      label_t _label;

      /// @brief This header size in bits.
      size_t _header_size;

      /// @brief The sender taskproxy.
      const TaskProxy_if_t * _sourceTaskproxy;

      /// @brief The destination task.
      const Task_if_t * _destinationTask;

      /// @brief The source node.
      const Node_t * _sourceNode;

      /// @brief The destination node.
      const Node_t * _destinationNode;

      //@}


      /// @name Simulation fields.
      //@{

      /// @brief The packet id.
      packet_id_t _id;

      /// @brief The transmission channel.
      const Channel_if_t * _channel;

      //@}

      /// @name To be removed.
      //@{

      /// @brief The validity flag.
      bool _is_valid;

      /// @brief Source node properties.
      node_properties_t  _sourceProperties;

      //@}

  };

} }



#endif

